% Copyright 2018 Tony Stillfjord
%
% This file is part of DREsplit.
% 
% DREsplit is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% DREsplit is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with DREsplit.  If not, see <http://www.gnu.org/licenses/>.

clear all
close('all')

%% Set up the problem
Nx = 50;
Nx2 = Nx^2;

% Finite-difference approximation of Laplacian on [0,1]^2
xx = linspace(0,1, Nx+2);
x = xx(2:end-1);
dx = 1/(Nx+1);
e = ones(Nx, 1);
A_1D = 1/dx^2*spdiags([e, -2*e, e], -1:1, Nx, Nx);

[X, Y] = meshgrid(x,x);
I_1D = eye(Nx, Nx);
A_2D = kron(A_1D, I_1D) + kron(I_1D, A_1D);
M = speye(Nx2, Nx2); % mass matrix = I in this case

% 3 inputs, control on the squares [j/4, j/4 + 1/8] x [j/4, j/4 + 1/8]
Nb = 3;
B = zeros(Nx2, Nb);
for j = 1:Nb
    B(:, j) = reshape((X > j/4) & (X < j/4 + 1/8) & (Y > j/4) & (Y < j/4 + 1/8), Nx2, 1);
end

% 1 output, average of all states
C = 1/Nx2 * ones(1, Nx2);
% LDLT-factorization of Q = C'C
LQ = C';
DQ = 1e0 * eye(size(C', 2));

% Time dependency through factors
alpha = @(t) 1 + 10*sin(2*pi*t); % A
mu = @(t) 2 + 7.1*sin(2*pi*t);   % M
dmu = @(t) 7.1*2*pi*cos(2*pi*t);
beta = @(t) 3 + cos(t);          % B
gamma = @(t) 1 - min(t, 1);      % C
% 
% alpha = @(t) 1; % A
% mu = @(t) 1;   % M
% dmu = @(t) 0;
% beta = @(t) 1;          % B
% gamma = @(t) 1;      % C


At = @(t) alpha(t) * A_2D;
Mt = @(t) mu(t) * M;
dMt = @(t) dmu(t) * M;
Bt = @(t) beta(t) * B;
LQt = @(t) gamma(t) * C';  

problem.A = At;
problem.M = Mt; 
problem.dM = dMt;
problem.B = Bt;
% problem.C = Ct;
problem.LQ = LQt;
problem.DQ = DQ;

% LDLT-factorization of initial condition P(0) = 0
L0 = zeros(Nx2, 1);
problem.L0 = L0;
problem.D0 = eye(size(L0, 2));

% R^{-1}, inverse of weighting factor for input in cost functional
% With R this small, the control has an impact. It also makes the quadratic
% term more significant.
problem.Rinv = 1e6*eye(size(problem.B, 2));

% Column compression tolerance
problem.CC_tol = eps;

%%%% Options for computing "matrix exponential actions"

% Alternative 5: use adaptive SDIRK (only choice for time-dependent problems)
problem.impODE.method = 'timevarying';
problem.impODE.tol = 1e-4;

%%%% Options for numerical quadrature applied to the integral term in the
% solution of the affine problem

% quadrature.type can be either 'gauss', 'clenshawcurtis' or 'equidistant'.
% 'gauss' means Gauss quadrature, which has optimal order for its number of
% points, but with no error estimate available.
% 'clenshawcurtis' uses a Clenshaw-Curtis formula
% 'adaptive' also uses a Clenshaw-Curtis formula, where the nodes may be 
% reused for error estimation. This is then used for increasing the 
% quadrature order until the requested tolerance quadrature.tol is met.
% The starting order is fixed to 4 and ignores quadrature.order.
% 'equidistant' uses equidistantly placed quadrature nodes and computes the
% corresponding weights by solving a Vandermonde system.
% If adaptive time-stepping is used with the option to recompute the
% integral factorization, equidistant nodes are always used and this option
% is ignored.

% problem.quadrature.type = 'gauss'; 
% problem.quadrature.order = 7; % optional; the methods choose if this is not supplied

problem.quadrature.type = 'adaptive'; % increase order until tolerance met
problem.quadrature.tol = 1e-4; % tolerance for above adaptivity

% Solve the systems for all quadrature points simultaneously. This will typically
% be faster due to BLAS-3 blocking effects, but will increase the memory
% requirements significantly.
problem.quadrature.blocking = false; 
problem.quadrature.LQ_size = [Nx2, size(C,1)]; % Specify dimensions of LQ(t) if blocking is used


% Integration interval
problem.t0 = 0;
problem.tend = 0.1;
% Numbers of time steps to use
problem.Nt = 2^4;  

% Output options 
problem.intermediates = true;       % Store intermediate steps 
problem.print_step = true;          % Print number of total steps and steps taken
problem.print_output = false;       % Print extra info every step

% Clean up and summarize arguments
disp(problem)

% Use Strang-splitting to solve the problem
problem.solver = 'Strang';
out = general_solver(problem);

% The solution approximation is given in factorized form by Ls*Ds*Ls'
% if problem.intermediates is true, then Ls{j}*Ds{j}*Ls{j}' is the 
% approximation at t(j)
% ms(j) contains the rank of the approximation at t(j)
% time is the main computation time


% Solutions can be efficiently compared in the Frobenius without forming 
% the dense matrices X = LDL' by using the function outerfrobnormdiff_LDLT
% as in the code snippet below. Here it is assumed that the above code was
% run several time with different options to generate LsEx, LsK, and LsIRK.
% (For the different matrix exponential action methods; exact, Krylov,
% IRK.)
% See also the example_error_framework.m file which shows how to use the 
% existing code for comparing solutions corresponding to different temporal 
% discretizations.

% 
% for j = 1:problem.Nt
% %     XIRK{j} = LsIRK{j}*DsIRK{j}*LsIRK{j}';
% %     XEx{j} = LsEx{j}*DsEx{j}*LsEx{j}';
% %     XK{j} = LsK{j}*DsK{j}*LsK{j}';
%     errsK(j) = outerfrobnormdiff_LDLT(LsK{j}, DsK{j}, LsEx{j}, DsEx{j}) / outerfrobnorm_LDLT(LsEx{j}, DsEx{j});
%     errsIRK(j) = outerfrobnormdiff_LDLT(LsIRK{j}, DsIRK{j}, LsEx{j}, DsEx{j}) / outerfrobnorm_LDLT(LsEx{j}, DsEx{j});
% end
% hs = problem.tend ./ (1:problem.Nt);
% 
% loglog(hs, errsK)
% hold on
% loglog(hs, errsIRK)
